
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/times.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <time.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <linux/soundcard.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>

/* SVGALIB header files */
#include <vga.h>
#include <vgagl.h>
#include <vgakeyboard.h>
#include <vgajoystick.h>

/* NLS Stuff */
#include <libintl.h>
#define _(String) gettext (String)
#define gettext_noop(String) (String)


#include "defines.h"

/* The soundIt library */
#include "soundIt.h"

/*
 * Do not modify the following file paths, they are defaults used
 * when the game is not installed.
 * The true ones are must be modified in the Makefile.
 */

#define LINBERTO_CONFIGFILE_DEFAULT "./data/config.dat"
#define LINBERTO_SCOREFILE_DEFAULT  "./data/highscores.dat"
#define LINBERTO_WORLDFILE_DEFAULT  "./data/default.wld"
#define LINBERTO_DATADIR_DEFAULT    "./data"
#define LINBERTO_SNDDIR_DEFAULT     "./data/sounds"

/* Linberto start X, Y coordinates */
#define LIN_START_X 141
#define LIN_START_Y  13

/* Bubble start X, Y coordinates */
#define BUB_START_X 140
#define BUB_START_Y 18

/* Flowerpot X, Y coordinates */
#define FLP_START_X 131
#define FLP_START_Y   0

/* Flowerpot kill X, Y coordinates (regarding the flowerpot X, Y) */
#define FLP_FPK_X  9
#define FLP_FPK_Y 35

/* Block X, Y coordinates */
#define BLOCK_X  38
#define BLOCK_Y  30

/* Maximun number of levels per stage */
#define MAX_LEVEL   4

/* Map index 
 * The are reversed!, and I am lazy to correct this :)
 */
#define MAP_X_NDX   7
#define MAP_Y_NDX  15

/* No know key pressed */
#define LIN_KEY_NONE 256

/* Setup defines */
#define LIVES_MAX           30
#define INVINCIBLE_OFF       0 /* Invincible */
#define INVINCIBLE_ON        1
#define SHOWTEXT_OFF         0
#define SHOWTEXT_ON          1
#define SOUND_OFF            0 /* Turn sound OFF and ON */
#define SOUND_ON             1
#define SOUND_SERVER_OFF     0 /* The sound server didn't started        */
#define SOUND_SERVER_ON      1 /* The sound server is active and running */
#define SOUND_CHANNELS       8
#define LIN_LANGUAGE_ENGLISH 0
#define LIN_LANGUAGE_SPANISH 1
#define LIN_LANGUAGE_ITALIAN 2
#define LIN_LANGUAGE_GERMAN  3
#define LAST_LANGUAGE        LIN_LANGUAGE_GERMAN
#define LIN_JOYSTICK_OFF     0
#define LIN_JOYSTICK_ON      1
#define BLOCK_PLAIN          0
#define BLOCK_TEXTURE        1
#define BLOCK_WIRE           2
#define BLOCK_SPHERE         3
#define BLOCK_MIN            BLOCK_PLAIN
#define BLOCK_MAX            BLOCK_SPHERE
#define ENEMY_DUMB           0
#define ENEMY_GENIUS         1
#define ENEMY_MIN            ENEMY_DUMB
#define ENEMY_MAX            ENEMY_GENIUS
#define SPEED_X01            0.1
#define SPEED_X1             1.0
#define SPEED_X9             9.0
#define SPEED_MIN            0.0
#define SPEED_MAX            SPEED_X9
#define DEFAULT_LANGUAGE     LIN_LANGUAGE_ENGLISH
#define DEFAULT_STAGE        0
#define DEFAULT_INVINCIBLE   INVINCIBLE_OFF
#define DEFAULT_SHOWTEXT     SHOWTEXT_ON
#define DEFAULT_SOUND        SOUND_ON
#define DEFAULT_LIVES        3
#define DEFAULT_BLOCKKIND    BLOCK_PLAIN
#define DEFAULT_ENEMYKIND    ENEMY_GENIUS
#define DEFAULT_SPEED        SPEED_X1
#define DEFAULT_JOYSTICK     LIN_JOYSTICK_OFF

/* Menu spacing (space for the next option in the menu (in pixels)). */
#define DEFAULT_MENU_SPACING 10

/* Does the actor appeared? */
#define ACTOR_INACTIVE  0
#define ACTOR_ACTIVE    1

/* The most used colors with this palette */
#define LIN_WHITE     51
#define LIN_BLACK      0
#define LIN_RED      181

/* Level colors */
#define WHITE_TOP     51
#define WHITE_LEFT    38
#define WHITE_RIGHT   26
#define BLUE_TOP      98 
#define BLUE_LEFT     78
#define BLUE_RIGHT    69 
#define GREEN_TOP    149
#define GREEN_LEFT   129
#define GREEN_RIGHT  122
#define RED_TOP      201
#define RED_LEFT     181
#define RED_RIGHT    172
#define YELLOW_TOP   248
#define YELLOW_LEFT  231
#define YELLOW_RIGHT 223

/* AI for moving the evils */
#define MOVE_UL    0 /* Up left               */
#define MOVE_UR    1 /* Up right              */
#define MOVE_DL    2 /* Down left             */
#define MOVE_DR    3 /* Down right            */
#define MOVE_NONE  4 /* Don't move            */
#define MOVE_START 5 /* Move to the beginning */

/* Waiting behavior */
#define WAIT_ESC_ESC   0
#define WAIT_ESC_ENTER 1

/* The largest number rand will return (same as INT_MAX).  */
#ifndef RAND_MAX
#define RAND_MAX        2147483647
#endif

#ifndef MIN
#define MIN( a, b) ((a) < (b) ? (a) : (b))
#endif

#ifndef MAX
#define MAX( a, b) ((a) > (b) ? (a) : (b))
#endif

/* Global variables */

struct level_colors {
       unsigned char ctop;
       unsigned char cleft;
       unsigned char cright;
};

struct sprite {
       int x, y;     /* X, Y position of sprite in screen  */
       int width;    /* width and height of the sprite     */
       int height;
       void *image;  /* The sprite image bytes             */
};

/* This struct is used to know the last used position of a
   sprite, to put it correctly in the next position        */
struct last_position {
       int lx, ly;   /* Last X, Y of a sprite             */
       int lw, lh;   /* Last Width and height of a sprite */
};

/* This is the struct for saving the score */
struct score_struct {
       char name[ 4];
       long score;
};

/* The header of the worlds file */
struct world_header {
       char magic[ 16]; /* Magic number (for id.)        */
#define WORLD_MAGIC  "LINBERTO WORLD"
       int number;      /* Number of stages in this file */
};

struct config_file {
       int cfg_language;           /* Language                       */
       int cfg_lives;              /* Number of lives                */
       int cfg_showtext;           /* Show text info?                */
       int cfg_invincible;         /* Is Linberto Invincible?        */
       int cfg_start_stage;        /* Starting Stage                 */
       int cfg_sound;              /* Sound ON/OFF                   */
       int cfg_blockkind;          /* Kind of block                  */
       int cfg_enemykind;          /* The kind of enemy intelligence */
       double cfg_speed;           /* Speed of enemies               */
       char cfg_world_path[ 1024]; /* filename for the world         */
       int cfg_joystick;           /* Joystick ON/OFF                */
};

/* The linked list structure */
struct ll_list {
        char *filename;
        int stages;
        struct ll_list *next;
        struct ll_list *prev;
};

/*
 * The main structure of the game.
 * Used to hold almost all the data for each actor. (Linberto, Bubble,
 * Flowerpot, etc).
 */
struct actor_struct {
       int mapx, mapy;          /* X & Y position on the map stage array */
       int active;              /* Does the actor appeared?              */
       struct sprite *sp;       /* Pointer to sprite structure           */
       struct last_position lp; /* Last position of the sprite           */
       clock_t start_time;      /* Place to save the current clock tick  */
                                /* used to know when to appear.          */

                                /* Time to wait to appear for each level */
                                /* (In clock ticks) see man 2 times      */
       clock_t wait_clockticks[ MAX_LEVEL + 1];
       void (* init_func)( int);  /* Function to initialize the actor      */
       void (* kill_func)( void); /* Function to put the sprite that kills */
                                  /* Linberto                              */
};

GraphicsContext *physicalscreen;
GraphicsContext *backscreen;
GraphicsContext *spritescreen;

/* note: extern variables are explained where defined, just use grep. */

extern struct actor_struct act_linberto;
extern struct actor_struct act_bubble;
extern struct actor_struct act_flowerpot;

extern int  stage;
extern int  level;
extern int  lives;
extern long score;

extern int bright;

extern int max_stages;

extern struct level_colors lcolors[];

extern char *openedfile;
extern char map_stage[ MAP_X_NDX][ MAP_Y_NDX];

extern int lin_scancode;
extern char keyttbl[ 128];

extern int send_sound_flag;

extern struct ll_list *ll_head;
extern struct ll_list *ll_tail;
extern struct ll_list *ll_current;
extern struct ll_list *ll_browse_curr;
extern struct ll_list *ll_browse_top;
extern int ll_qty;
extern int ll_browse_pos;

/* Setup variables */
extern struct config_file opt;

extern char world_filename[ 4096];

extern int menu_spacing;
extern void (* menu_function)( int);

extern char *nls_restore;

/*
 * Sprites
 */

/* Linberto */
extern unsigned char lbertosp[];
extern struct sprite sprite_linberto;

/* The evil ones */
extern unsigned char bubble[];
extern unsigned char bubble_trap[];
extern unsigned char flowerpot[];
extern unsigned char pot_kill[];
extern struct sprite sprite_bubble;
extern struct sprite sprite_bubble_trap;
extern struct sprite sprite_flowerpot;
extern struct sprite sprite_pot_kill;
extern unsigned char title[];

/* sound enum */
enum sound {
        /* Music */
        SND_WELCOME, /* Welcome/main menu music    */
        SND_NONE,    /* No music                   */
        SND_BKGDM,   /* Background music           */

        /* Sound FX */
        SND_DEATH,   /* Linberto died              */
        SND_JUMP,    /* Linberto jumped            */
        SND_SPLAT,   /* Flowerpot killed Linberto  */
        SND_FPJP,    /* Flowerpot jumped           */
        SND_BUBJP,   /* Bubble jumped              */
        SND_CHANGE,  /* Change option              */
        SND_FALL,    /* Linberto fall              */
        SND_MENU,    /* Menu move sound            */
        SND_BUBK,    /* Bubble killed Linberto     */
        SND_BLKD,    /* Block Destroy              */
        SND_BLKF,    /* Block Fill                 */
        SND_BLKM,    /* Block Move                 */
        SND_BLKN,    /* Block Next/Previous...     */
        SND_MAX      /* Maximum number of sounds   */
};

/* Function prototypes */

/* draw.c */
void draw_berto_board( int ndx);
void draw_block( int x, int y, struct level_colors lc);
void draw_cube( int x, int y, struct level_colors lc);
void draw_board_top( int sx, int sy, int sidecolor);
void draw_board_sideleft( int sx, int sy, int sidecolor);
void draw_board_sideright( int sx, int sy, int sidecolor);
void draw_mini_board( int x, int y);
void draw_mini_cube( int x, int y, struct level_colors lc);
void draw_mini_board_top( int sx, int sy, int sidecolor);
void draw_mini_board_sideleft( int sx, int sy, int sidecolor);
void draw_mini_board_sideright( int sx, int sy, int sidecolor);
void draw_wire_cube( int x, int y, struct level_colors lc);
void draw_texture_cube( int x, int y, struct level_colors lc);
void draw_texture_top( int sx, int sy, int sidecolor);
void draw_texture_sideleft( int sx, int sy, int sidecolor);
void draw_texture_sideright( int sx, int sy, int sidecolor);
void draw_bar( int x, int y, int len, int pos, int col);
void draw_sphere( int x, int y, struct level_colors lc);
void draw_circle( int x, int y, int r, int c);
void draw_rectangle( int x, int y, int w, int h, int c);

/* game.c */
void game_loop( void);
void game_init_actors( void);
void game_init_linberto( int active_status);
void game_init_bubble( int active_status);
void game_init_flowerpot( int active_status);
void game_init_map_stage( char *filename);
void game_move_beginning( void);
void move_sprite( void);
void move_sprite_downleft( void);
void move_sprite_downright( void);
void move_sprite_upleft( void);
void move_sprite_upright( void);
void game_linberto_died( void);
int game_checkpos( void);
int game_check_complete( void);
void game_draw_block( void);
void game_show_text( void);
int game_check_evils( struct actor_struct *act);
int game_where_to_move( struct actor_struct *act);
void game_linberto_bubble_trapped( void);
void game_linberto_flowerpot_kill( void);
void game_delete_sprite( struct actor_struct *act);
void game_show_stage( void);
void game_set_evil_timers( struct actor_struct *act);
void game_win( void);
void game_pause( void);
void game_sound( void);
void game_over( void);

/* help.c */
void help_show_instructions( void);

/* init.c */
void init_svgalib( void);
void init_save_config( void);
void init_load_stage( char *filename);

/* joy.c */
void joy_init( void);
int joy_moved( void);
int joy_button_enter( void);
int joy_button_any( void);
int joy_can_get_sample( void);
int joy_equal( int value);
int joy_more( int value);
int joy_less( int value);

/* key.c */
void key_handler( int scancode, int press);
void key_wait_for_enter( void);
int key_find_char( char c, int sum);

/* llist.c */
void ll_init( void);
struct ll_list *ll_alloc( char *filename);
void ll_add( char *filename, int stages);
void ll_free( void);
void ll_show( void);
int ll_browse_next( int nitems);
int ll_browse_back( int nitems);

/* misc.c */
void exit_game( int status);
void *lin_calloc( int size);
void press_enter_str( void);
void misc_calculate_minmax( struct actor_struct *act, int *x, int *y, int *w, int *h);
void misc_save_last_position( struct actor_struct *act);
void check_if_file_exist( char *filename);
FILE *linberto_openfile( char *name1, char *name2, char *mode);
int menu( int x, int y, int len, int pos, int maxpos, int quitpos);
int loopbar( int pos, int maxpos, int x, int y, int len);
int check_world_magic( char *filename, int *stages);
void lin_chdir( char *dir);
void lin_title_line( void);
int lin_inputstr( char *title, char *str, int qty);
void grf_printf( int x, int y, char *fmt, ...);
int misc_wait( int ticks, int behavior);

/* nls.c */
void nls_init( void);
void nls_change_language( void);
void nls_exit( void);
int nls_get_default_language( void);

/* setup.c */
void setup( void);
void setup_block_change_drawcubes( int pos);

/* palette.c */
void palette_init( void);
void palette_set_brightness( void);

/* screen.c */
void screen_center_message( char *msg, int col);
void screen_center_text( char *msg, int y);
void screen_shot( void);

/* score.c */
void score_show( void);
void score_high_check( void);

/* signals.c */
void init_signals( void);
void handler( int signalnumber);

/* sound.c */
void sound_init( void);
void sound_off( void);
void sound_send( int snd_number);
void sound_music( int snd_number);
void sound_shutdown( void);

/* sprites.c */
void sprite_put( struct sprite *sp);
void sprite_put_mask( struct sprite *sp);

/* world.c */
void world_editor( void);
void world_select( void);
void world_edit_menu( void);
void world_navigate_names( void);
int world_edit( void);
int world_delete( void);
int world_check_exist( void);

